from abc import ABC, abstractmethod
from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING
from uuid import uuid4

from pydantic import BaseModel, Field

from hatchet_sdk.config import ClientConfig
from hatchet_sdk.contracts.v1.shared.condition_pb2 import Action as ProtoAction
from hatchet_sdk.contracts.v1.shared.condition_pb2 import (
    BaseMatchCondition,
    ParentOverrideMatchCondition,
    SleepMatchCondition,
    UserEventMatchCondition,
)
from hatchet_sdk.utils.proto_enums import convert_python_enum_to_proto
from hatchet_sdk.utils.timedelta_to_expression import Duration, timedelta_to_expr

if TYPE_CHECKING:
    from hatchet_sdk.runnables.task import Task
    from hatchet_sdk.runnables.types import R, TWorkflowInput


def generate_or_group_id() -> str:
    return str(uuid4())


class Action(Enum):
    CREATE = 0
    QUEUE = 1
    CANCEL = 2
    SKIP = 3


class BaseCondition(BaseModel):
    readable_data_key: str
    action: Action | None = None
    or_group_id: str = Field(default_factory=generate_or_group_id)
    expression: str | None = None

    def to_proto(self) -> BaseMatchCondition:
        return BaseMatchCondition(
            readable_data_key=self.readable_data_key,
            action=convert_python_enum_to_proto(self.action, ProtoAction),  # type: ignore[arg-type]
            or_group_id=self.or_group_id,
            expression=self.expression,
        )


class Condition(ABC):
    def __init__(self, base: BaseCondition):
        self.base = base

    @abstractmethod
    def to_proto(
        self, config: ClientConfig
    ) -> UserEventMatchCondition | ParentOverrideMatchCondition | SleepMatchCondition:
        pass


class SleepCondition(Condition):
    def __init__(
        self, duration: Duration, readable_data_key: str | None = None
    ) -> None:
        super().__init__(
            BaseCondition(
                readable_data_key=readable_data_key
                or f"sleep:{timedelta_to_expr(duration)}",
            )
        )

        self.duration = duration

    def to_proto(self, config: ClientConfig) -> SleepMatchCondition:
        return SleepMatchCondition(
            base=self.base.to_proto(),
            sleep_for=timedelta_to_expr(self.duration),
        )


class UserEventCondition(Condition):
    def __init__(
        self,
        event_key: str,
        expression: str | None = None,
        readable_data_key: str | None = None,
    ) -> None:
        super().__init__(
            BaseCondition(
                readable_data_key=readable_data_key or event_key,
                expression=expression,
            )
        )

        self.event_key = event_key
        self.expression = expression

    def to_proto(self, config: ClientConfig) -> UserEventMatchCondition:
        return UserEventMatchCondition(
            base=self.base.to_proto(),
            user_event_key=config.apply_namespace(self.event_key),
        )


class ParentCondition(Condition):
    def __init__(
        self,
        parent: "Task[TWorkflowInput, R]",
        expression: str | None = None,
        readable_data_key: str | None = None,
    ) -> None:
        super().__init__(
            BaseCondition(
                readable_data_key=readable_data_key
                or (
                    parent.name
                    + (f":{expression}" if expression else "")
                    + ":"
                    + datetime.now(tz=timezone.utc).isoformat()
                ),
                expression=expression,
            )
        )

        self.parent = parent

    def to_proto(self, config: ClientConfig) -> ParentOverrideMatchCondition:
        return ParentOverrideMatchCondition(
            base=self.base.to_proto(),
            parent_readable_id=self.parent.name,
        )


class OrGroup:
    def __init__(self, conditions: list[Condition]) -> None:
        self.or_group_id = generate_or_group_id()
        self.conditions = conditions


def or_(*conditions: Condition) -> OrGroup:
    return OrGroup(conditions=list(conditions))


def flatten_conditions(conditions: list[Condition | OrGroup]) -> list[Condition]:
    flattened: list[Condition] = []

    for condition in conditions:
        if isinstance(condition, OrGroup):
            for or_condition in condition.conditions:
                or_condition.base.or_group_id = condition.or_group_id

            flattened.extend(condition.conditions)
        else:
            flattened.append(condition)

    return flattened



from hatchet_sdk.serde.pydantic import (
    HATCHET_PYDANTIC_SENTINEL,
    is_in_hatchet_serialization_context,
)

__all__ = [
    "HATCHET_PYDANTIC_SENTINEL",
    "is_in_hatchet_serialization_context",
]



import os
from typing import Literal, cast, overload

import grpc

from hatchet_sdk.config import ClientConfig


@overload
def new_conn(config: ClientConfig, aio: Literal[False]) -> grpc.Channel: ...


@overload
def new_conn(config: ClientConfig, aio: Literal[True]) -> grpc.aio.Channel: ...


def new_conn(config: ClientConfig, aio: bool) -> grpc.Channel | grpc.aio.Channel:
    credentials: grpc.ChannelCredentials | None = None

    # load channel credentials
    if config.tls_config.strategy == "tls":
        root: bytes | None = None

        if config.tls_config.root_ca_file:
            with open(config.tls_config.root_ca_file, "rb") as f:
                root = f.read()

        credentials = grpc.ssl_channel_credentials(root_certificates=root)
    elif config.tls_config.strategy == "mtls":
        assert config.tls_config.root_ca_file
        assert config.tls_config.key_file
        assert config.tls_config.cert_file

        with open(config.tls_config.root_ca_file, "rb") as f:
            root = f.read()

        with open(config.tls_config.key_file, "rb") as f:
            private_key = f.read()

        with open(config.tls_config.cert_file, "rb") as f:
            certificate_chain = f.read()

        credentials = grpc.ssl_channel_credentials(
            root_certificates=root,
            private_key=private_key,
            certificate_chain=certificate_chain,
        )

    start = grpc if not aio else grpc.aio

    channel_options: list[tuple[str, str | int]] = [
        ("grpc.max_send_message_length", config.grpc_max_send_message_length),
        ("grpc.max_receive_message_length", config.grpc_max_recv_message_length),
        ("grpc.keepalive_time_ms", 10 * 1000),
        ("grpc.keepalive_timeout_ms", 60 * 1000),
        ("grpc.client_idle_timeout_ms", 60 * 1000),
        ("grpc.http2.max_pings_without_data", 0),
        ("grpc.keepalive_permit_without_calls", 1),
        ("grpc.default_compression_algorithm", grpc.Compression.Gzip),
    ]

    # Set environment variable to disable fork support. Reference: https://github.com/grpc/grpc/issues/28557
    # When steps execute via os.fork, we see `TSI_DATA_CORRUPTED` errors.
    # needs to be the string "True" or "False"
    os.environ["GRPC_ENABLE_FORK_SUPPORT"] = str(config.grpc_enable_fork_support)

    if config.grpc_enable_fork_support:
        # See discussion: https://github.com/hatchet-dev/hatchet/pull/2057#discussion_r2243233357
        os.environ["GRPC_POLL_STRATEGY"] = "poll"

    if config.tls_config.strategy == "none":
        conn = start.insecure_channel(
            target=config.host_port,
            options=channel_options,
        )
    else:
        channel_options.append(
            ("grpc.ssl_target_name_override", config.tls_config.server_name)
        )

        conn = start.secure_channel(
            target=config.host_port,
            credentials=credentials,
            options=channel_options,
        )

    return cast(
        grpc.Channel | grpc.aio.Channel,
        conn,
    )



from enum import Enum

from pydantic import BaseModel, model_validator

from hatchet_sdk.contracts.v1.workflows_pb2 import CreateTaskRateLimit


class RateLimitDuration(str, Enum):
    SECOND = "SECOND"
    MINUTE = "MINUTE"
    HOUR = "HOUR"
    DAY = "DAY"
    WEEK = "WEEK"
    MONTH = "MONTH"
    YEAR = "YEAR"


class RateLimit(BaseModel):
    """
    Represents a rate limit configuration for a step in a workflow.

    This class allows for both static and dynamic rate limiting based on various parameters.
    It supports both simple integer values and Common Expression Language (CEL) expressions
    for dynamic evaluation.

    Attributes:
        static_key (str, optional): A static key for rate limiting.
        dynamic_key (str, optional): A CEL expression for dynamic key evaluation.
        units (int or str, default=1): The number of units or a CEL expression for dynamic unit calculation.
        limit (int or str, optional): The rate limit value or a CEL expression for dynamic limit calculation.
        duration (str, default=RateLimitDuration.MINUTE): The window duration of the rate limit.

    Usage:
        1. Static rate limit:
           rate_limit = RateLimit(static_key="external-api", units=100)
           > NOTE: if you want to use a static key, you must first put the rate limit: hatchet.admin.put_rate_limit("external-api", 200, RateLimitDuration.SECOND)

        2. Dynamic rate limit with CEL expressions:
           rate_limit = RateLimit(
               dynamic_key="input.user_id",
               units="input.units",
               limit="input.limit * input.user_tier"
           )

    Note:
        - Either static_key or dynamic_key must be set, but not both.
        - When using dynamic_key, limit must also be set.
        - CEL expressions are validated upon instantiation.

    Raises:
        ValueError: If invalid combinations of attributes are provided or if CEL expressions are invalid.
    """

    static_key: str | None = None
    dynamic_key: str | None = None
    units: str | int = 1
    limit: int | str | None = None
    duration: RateLimitDuration = RateLimitDuration.MINUTE

    @model_validator(mode="after")
    def validate_rate_limit(self) -> "RateLimit":
        if self.dynamic_key and self.static_key:
            raise ValueError("Cannot have both static key and dynamic key set")

        if self.limit and not isinstance(self.limit, int | str):
            raise ValueError(f"Invalid limit value: {self.limit}")

        if self.dynamic_key and not self.limit:
            raise ValueError("CEL based keys requires limit to be set")

        if self.limit is None:
            self.limit = -1

        return self

    def to_proto(self) -> CreateTaskRateLimit:
        key = self.static_key
        key_expression = self.dynamic_key

        key = self.static_key or self.dynamic_key

        units = self.units if isinstance(self.units, int) else None
        units_expression = None if isinstance(self.units, int) else self.units

        limit_expression = None if not self.limit else str(self.limit)

        return CreateTaskRateLimit(
            key=key,
            key_expr=key_expression,
            units=units,
            units_expr=units_expression,
            limit_values_expr=limit_expression,
            duration=self.duration,
        )



import asyncio

from pydantic import BaseModel, Field, field_validator

from hatchet_sdk.clients.rest.api.workflow_api import WorkflowApi
from hatchet_sdk.clients.rest.api.workflow_run_api import WorkflowRunApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.create_cron_workflow_trigger_request import (
    CreateCronWorkflowTriggerRequest,
)
from hatchet_sdk.clients.rest.models.cron_workflows import CronWorkflows
from hatchet_sdk.clients.rest.models.cron_workflows_list import CronWorkflowsList
from hatchet_sdk.clients.rest.models.cron_workflows_order_by_field import (
    CronWorkflowsOrderByField,
)
from hatchet_sdk.clients.rest.models.workflow_run_order_by_direction import (
    WorkflowRunOrderByDirection,
)
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import (
    BaseRestClient,
    maybe_additional_metadata_to_kv,
)
from hatchet_sdk.utils.typing import JSONSerializableMapping


class CreateCronTriggerConfig(BaseModel):
    """
    Schema for creating a workflow run triggered by a cron.
    """

    expression: str
    input: JSONSerializableMapping = Field(default_factory=dict)
    additional_metadata: JSONSerializableMapping = Field(default_factory=dict)

    @field_validator("expression")
    @classmethod
    def validate_cron_expression(cls, v: str) -> str:
        """
        Validates the cron expression to ensure it adheres to the expected format.

        :param v: The cron expression to validate.

        :raises ValueError: If the expression is invalid

        :return: The validated cron expression.
        """
        if not v:
            raise ValueError("Cron expression is required")

        ## allow cron aliases
        if (alias := v.strip()) in {
            "@yearly",
            "@annually",
            "@monthly",
            "@weekly",
            "@daily",
            "@hourly",
        }:
            return alias

        parts = v.split()
        if len(parts) != 5:
            raise ValueError(
                "Cron expression must have 5 parts: minute hour day month weekday"
            )

        for part in parts:
            if not (
                part == "*"
                or part.replace("*/", "").replace("-", "").replace(",", "").isdigit()
            ):
                raise ValueError(f"Invalid cron expression part: {part}")

        return v


class CronClient(BaseRestClient):
    """
    The cron client is a client for managing cron workflows within Hatchet.
    """

    def _wra(self, client: ApiClient) -> WorkflowRunApi:
        return WorkflowRunApi(client)

    def _wa(self, client: ApiClient) -> WorkflowApi:
        return WorkflowApi(client)

    def create(
        self,
        workflow_name: str,
        cron_name: str,
        expression: str,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
        priority: int | None = None,
    ) -> CronWorkflows:
        """
        Create a new workflow cron trigger.

        :param workflow_name: The name of the workflow to trigger.
        :param cron_name: The name of the cron trigger.
        :param expression: The cron expression defining the schedule.
        :param input: The input data for the cron workflow.
        :param additional_metadata: Additional metadata associated with the cron trigger.
        :param priority: The priority of the cron workflow trigger.

        :return: The created cron workflow instance.
        """
        validated_input = CreateCronTriggerConfig(
            expression=expression, input=input, additional_metadata=additional_metadata
        )

        with self.client() as client:
            return self._wra(client).cron_workflow_trigger_create(
                tenant=self.client_config.tenant_id,
                workflow=self.client_config.apply_namespace(workflow_name),
                create_cron_workflow_trigger_request=CreateCronWorkflowTriggerRequest(
                    cronName=cron_name,
                    cronExpression=validated_input.expression,
                    input=validated_input.input,
                    additionalMetadata=validated_input.additional_metadata,
                    priority=priority,
                ),
            )

    async def aio_create(
        self,
        workflow_name: str,
        cron_name: str,
        expression: str,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
        priority: int | None = None,
    ) -> CronWorkflows:
        """
        Create a new workflow cron trigger.

        :param workflow_name: The name of the workflow to trigger.
        :param cron_name: The name of the cron trigger.
        :param expression: The cron expression defining the schedule.
        :param input: The input data for the cron workflow.
        :param additional_metadata: Additional metadata associated with the cron trigger.
        :param priority: The priority of the cron workflow trigger.

        :return: The created cron workflow instance.
        """
        return await asyncio.to_thread(
            self.create,
            workflow_name,
            cron_name,
            expression,
            input,
            additional_metadata,
            priority,
        )

    def delete(self, cron_id: str) -> None:
        """
        Delete a workflow cron trigger.

        :param cron_id: The ID of the cron trigger to delete.
        :return: None
        """
        with self.client() as client:
            self._wa(client).workflow_cron_delete(
                tenant=self.client_config.tenant_id, cron_workflow=str(cron_id)
            )

    async def aio_delete(self, cron_id: str) -> None:
        """
        Delete a workflow cron trigger.

        :param cron_id: The ID of the cron trigger to delete.
        :return: None
        """
        return await asyncio.to_thread(self.delete, cron_id)

    async def aio_list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
        order_by_field: CronWorkflowsOrderByField | None = None,
        order_by_direction: WorkflowRunOrderByDirection | None = None,
        workflow_name: str | None = None,
        cron_name: str | None = None,
    ) -> CronWorkflowsList:
        """
        Retrieve a list of all workflow cron triggers matching the criteria.

        :param offset: The offset to start the list from.
        :param limit: The maximum number of items to return.
        :param workflow_id: The ID of the workflow to filter by.
        :param additional_metadata: Filter by additional metadata keys.
        :param order_by_field: The field to order the list by.
        :param order_by_direction: The direction to order the list by.
        :param workflow_name: The name of the workflow to filter by.
        :param cron_name: The name of the cron trigger to filter by.

        :return: A list of cron workflows.
        """
        return await asyncio.to_thread(
            self.list,
            offset=offset,
            limit=limit,
            workflow_id=workflow_id,
            additional_metadata=additional_metadata,
            order_by_field=order_by_field,
            order_by_direction=order_by_direction,
            workflow_name=workflow_name,
            cron_name=cron_name,
        )

    def list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
        order_by_field: CronWorkflowsOrderByField | None = None,
        order_by_direction: WorkflowRunOrderByDirection | None = None,
        workflow_name: str | None = None,
        cron_name: str | None = None,
    ) -> CronWorkflowsList:
        """
        Retrieve a list of all workflow cron triggers matching the criteria.

        :param offset: The offset to start the list from.
        :param limit: The maximum number of items to return.
        :param workflow_id: The ID of the workflow to filter by.
        :param additional_metadata: Filter by additional metadata keys.
        :param order_by_field: The field to order the list by.
        :param order_by_direction: The direction to order the list by.
        :param workflow_name: The name of the workflow to filter by.
        :param cron_name: The name of the cron trigger to filter by.

        :return: A list of cron workflows.
        """
        with self.client() as client:
            cron_workflow_list = tenacity_retry(
                self._wa(client).cron_workflow_list,
                self.client_config.tenacity,
            )
            return cron_workflow_list(
                tenant=self.client_config.tenant_id,
                offset=offset,
                limit=limit,
                workflow_id=workflow_id,
                additional_metadata=maybe_additional_metadata_to_kv(
                    additional_metadata
                ),
                order_by_field=order_by_field,
                order_by_direction=order_by_direction,
                workflow_name=workflow_name,
                cron_name=cron_name,
            )

    def get(self, cron_id: str) -> CronWorkflows:
        """
        Retrieve a specific workflow cron trigger by ID.

        :param cron_id: The cron trigger ID or CronWorkflows instance to retrieve.
        :return: The requested cron workflow instance.
        """
        with self.client() as client:
            workflow_cron_get = tenacity_retry(
                self._wa(client).workflow_cron_get, self.client_config.tenacity
            )
            return workflow_cron_get(
                tenant=self.client_config.tenant_id, cron_workflow=str(cron_id)
            )

    async def aio_get(self, cron_id: str) -> CronWorkflows:
        """
        Retrieve a specific workflow cron trigger by ID.

        :param cron_id: The cron trigger ID or CronWorkflows instance to retrieve.
        :return: The requested cron workflow instance.
        """
        return await asyncio.to_thread(self.get, cron_id)
